"""
Test for FaML

by Lukas Chrostowski 2024

"""

def test_FaML_two():
    '''
    --- Simple MZI, tested using Facet-Attached Micro Lenses (FaML) ---
    
    by Lukas Chrostowski, 2024
    
    Example simple script to
    - choose the fabrication technology provided by Applied Nanotools,  using silicon nitride (SiN) waveguides
    - use the SiEPIC-EBeam-PDK technology
    - using KLayout and SiEPIC-Tools, with function including connect_pins_with_waveguide and connect_cell
    - create a new layout with a top cell, limited a design area of 1000 microns wide by 410 microns high.
    - create two Mach-Zehnder Interferometer (MZI) circuits, and one loopback for calibration
    One Mach-Zehnder has a small path length difference, while the other uses a very long spiral.
    - export to OASIS for submission to fabrication
    - display the layout in KLayout using KLive
    
    Test plan
    - count lenses from the bottom up (bottom is 1, top is 6, in this design)
    - laser input on bottom lens (1), detector on second (2), for alignment
    - MZI1: laser on 3, detector on 4, sweep
    - MZI2: laser on 5, detector on 6, sweep
    

    Use instructions:

    Run in Python, e.g., VSCode

    pip install required packages:
    - klayout, SiEPIC, siepic_ebeam_pdk, numpy

    '''

    designer_name = 'LukasChrostowski'
    top_cell_name = 'EBeam_%s_MZI2_FaML' % designer_name
    export_type = 'static'  # static: for fabrication, PCell: include PCells in file
    #export_type = 'PCell'  # static: for fabrication, PCell: include PCells in file

    import pya

    import SiEPIC
    from SiEPIC._globals import Python_Env
    from SiEPIC.scripts import connect_cell, connect_pins_with_waveguide, zoom_out, export_layout
    from SiEPIC.utils.layout import new_layout, floorplan, FaML_two
    from SiEPIC.extend import to_itype
    from SiEPIC.verification import layout_check
    
    import os

    if Python_Env == 'Script':
        # For external Python mode, when installed using pip install siepic_ebeam_pdk
        import siepic_ebeam_pdk

    print('EBeam_LukasChrostowski_MZI2 layout script')
    
    tech_name = 'EBeam'

    from packaging import version
    if version.parse(SiEPIC.__version__) < version.parse("0.5.4"):
        raise Exception("Errors", "This example requires SiEPIC-Tools version 0.5.4 or greater.")

    '''
    Create a new layout using the EBeam technology,
    with a top cell
    and Draw the floor plan
    '''    
    cell, ly = new_layout(tech_name, top_cell_name, GUI=True, overwrite = True)
    floorplan(cell, 1000e3, 244e3)

    waveguide_type1='SiN Strip TE 1550 nm, w=750 nm'
    waveguide_type_delay='SiN routing TE 1550 nm (compound waveguide)'

    # Load cells from library
    cell_ebeam_y = ly.create_cell('ANT_MMI_1x2_te1550_3dB_BB',  'EBeam-SiN')
    cell_ebeam_delay = ly.create_cell('spiral_paperclip', 'EBeam_Beta',
                                    {'waveguide_type':waveguide_type_delay,
                                    'length':319,
                                    'loops':8,
                                    'flatten':True})

    #######################
    # Circuit #2 – MZI, with a very long delay line
    #######################
    # draw two edge couplers for facet-attached micro-lenses
    inst_faml = FaML_two(cell, 
            label = "opt_in_TE_1550_FaML_mzi2_%s" % designer_name,
            cell_params = None
            )    
    # Y branches:
    instY2 = connect_cell(inst_faml[0], 'opt1', cell_ebeam_y, 'pin1')
    instY1 = connect_cell(inst_faml[1], 'opt1', cell_ebeam_y, 'pin1')
    # Spiral:
    instSpiral = connect_cell(instY2, 'pin2', cell_ebeam_delay, 'optA')
    instSpiral.transform(pya.Trans(110e3,50e3))
    # Waveguides:
    connect_pins_with_waveguide(instY1, 'pin2', instY2, 'pin3', waveguide_type=waveguide_type1)
    connect_pins_with_waveguide(instY2, 'pin2', instSpiral, 'optA', waveguide_type=waveguide_type1)
    connect_pins_with_waveguide(instY1, 'pin3', instSpiral, 'optB', waveguide_type=waveguide_type1,turtle_A=[50,90])

    # Export for fabrication, removing PCells
    path = os.path.dirname(os.path.realpath(__file__))
    filename, extension = os.path.splitext(os.path.basename(__file__))
    if export_type == 'static':
        file_out = export_layout(cell, path, filename, relative_path = '..', format='oas', screenshot=True)
    else:
        file_out = os.path.join(path,'..',filename+'.oas')
        ly.write(file_out)

    # Verify
    file_lyrdb = os.path.join(path,filename+'.lyrdb')
    num_errors = layout_check(cell = cell, verbose=False, GUI=True, file_rdb=file_lyrdb)
    print('Number of errors: %s' % num_errors)
    if num_errors > 0:
        raise Exception ('Errors found in test_FaML_two')

if __name__ == "__main__":
    test_FaML_two()



# $description: Mach-Zehnder Interferometer (EBeam)
# $show-in-menu
# $group-name: Examples_EBeam
# $menu-path: siepic_menu.exlayout.begin

'''
--- Simple MZI ---

by Lukas Chrostowski, 2020-2024

Example simple script to
 - create a new layout with a top cell
 - create an MZI
 - export to OASIS for submission to fabrication

using SiEPIC-Tools function including connect_pins_with_waveguide and connect_cell

usage:
 - run this script in Python
'''

def MZI():
        
    designer_name = 'Example'
    top_cell_name = 'EBeam_%s_MZI' % designer_name

    import pya
    from pya import Trans, CellInstArray, Text

    import SiEPIC
    from SiEPIC._globals import Python_Env
    from SiEPIC.scripts import connect_cell, connect_pins_with_waveguide, zoom_out, export_layout
    from SiEPIC.utils.layout import new_layout, floorplan
    from SiEPIC.extend import to_itype

    import os

    if Python_Env == 'Script':
        try:
            # For external Python mode, when installed using pip install siepic_ebeam_pdk
            import siepic_ebeam_pdk
        except:
            # Load the PDK from a folder, e.g, GitHub, when running externally from the KLayout Application
            import os, sys
            path_GitHub = os.path.expanduser('~/Documents/GitHub/')
            sys.path.append(os.path.join(path_GitHub, 'SiEPIC_EBeam_PDK/klayout'))
            import siepic_ebeam_pdk

    tech_name = 'EBeam'

    from packaging import version
    if version.parse(SiEPIC.__version__) < version.parse('0.5.1'):
        raise Exception("Errors", "This example requires SiEPIC-Tools version 0.5.1 or greater.")

    '''
    Create a new layout using the EBeam technology,
    with a top cell
    and Draw the floor plan
    '''    
    cell, ly = new_layout(tech_name, top_cell_name, GUI=True, overwrite = True)
    floorplan(cell, 605e3, 410e3)  # units: nanometers

    dbu = ly.dbu

    waveguide_type='Strip TE 1550 nm, w=500 nm'
    waveguide_type_delay='Si routing TE 1550 nm (compound waveguide)'

    # Load cells from library
    if version.parse(SiEPIC.__version__) < version.parse('0.5.19'):
        cell_ebeam_gc = ly.create_cell('GC_TE_1550_8degOxide_BB', tech_name)
        cell_ebeam_y = ly.create_cell('ebeam_y_1550', tech_name)
        cell_ebeam_y_dream = ly.create_cell('ebeam_dream_splitter_1x2_te1550_BB', 'EBeam-Dream')
    else:
        cell_ebeam_gc = ly.create_cell2('GC_TE_1550_8degOxide_BB', tech_name)
        cell_ebeam_y = ly.create_cell2('ebeam_y_1550', tech_name)
        cell_ebeam_y_dream = ly.create_cell2('ebeam_dream_splitter_1x2_te1550_BB', 'EBeam-Dream')

    # grating couplers, place at absolute positions
    x,y = 60000, 15000
    t = pya.Trans(Trans.R0,x,y)
    instGC1 = cell.insert(pya.CellInstArray(cell_ebeam_gc.cell_index(), t))
    t = pya.Trans(Trans.R0,x,y+127000)
    instGC2 = cell.insert(pya.CellInstArray(cell_ebeam_gc.cell_index(), t))

    # automated test label
    text = pya.Text ("opt_in_TE_1550_device_%s_MZI1" % designer_name, t)
    cell.shapes(ly.layer(ly.TECHNOLOGY['Text'])).insert(text).text_size = 5/dbu

    # Y branches:
    # Version 1: place it at an absolute position:
    t = pya.Trans.from_s('r0 %s, %s' % (x+20000,y))
    instY1 = cell.insert(pya.CellInstArray(cell_ebeam_y.cell_index(), t))

    # Version 2: attach it to an existing component, then move relative
    instY2 = connect_cell(instGC2, 'opt1', cell_ebeam_y, 'opt1')
    instY2.transform(Trans(20000,-10000))

    # Waveguides:

    connect_pins_with_waveguide(instGC1, 'opt1', instY1, 'opt1', waveguide_type=waveguide_type)
    connect_pins_with_waveguide(instGC2, 'opt1', instY2, 'opt1', waveguide_type=waveguide_type)
    inst_wg1 = connect_pins_with_waveguide(instY1, 'opt2', instY2, 'opt3', waveguide_type=waveguide_type)
    inst_wg2 = connect_pins_with_waveguide(instY1, 'opt3', instY2, 'opt2', waveguide_type=waveguide_type,turtle_B=[25,-90])

    # determine path length difference for the MZI

    # Method 1
    def inst_spice_param(inst, param):
        '''
        Input:
            inst: an instance of a cell
            param: which spice parameter you want to query, e.g., param="wg_length"
                the cell needs to have a Text layer containing "spice_param:", e.g., "spice_param:wg_length=100.000 width=0.5"
        Output: 
            the numerical value of the parameter, in microns
        Approach:
            This function scans the GDS layout for a Text label ("spice_param").
            It should work with GDS files create by any tool, 
            as long as it contains the required Text label.
            Alternatively, this could be implemented by querying the PCell directly, but
            this would only work while the script is running.
        '''    
        # get "param" from instance
        params1 = inst.cell.find_components(cell_selected=inst.cell)[0].params
        param_val = dict([p.split('=') for p in params1.split(' ')])[param]
        return float(param_val)*1e6
        
    wg_length1 = inst_spice_param(inst_wg1, 'wg_length')
    wg_length2 = inst_spice_param(inst_wg2, 'wg_length')
    print('Length difference: %s; lengths: %s, %s' % (abs(wg_length2-wg_length1), wg_length2, wg_length1))

    # Method 2
    def wg_length_EBeam(inst):
        '''
        Input: 
            inst: an instance of a "EBeam.Waveguide" PCell
        Output: 
            the waveguide length, in microns
        Approach:
            Use the display title for the cell
            only works inside a scripted layout
        '''
        return float(inst.cell.display_title().split('_')[-1])

    wg_length1 = wg_length_EBeam(inst_wg1)
    wg_length2 = wg_length_EBeam(inst_wg2)
    print('Length difference: %s; lengths: %s, %s' % (abs(wg_length2-wg_length1), wg_length2, wg_length1))


    # 2nd MZI using Dream Photonics 1x2 splitter
    # grating couplers, place at absolute positions
    x,y = 180000, 15000
    t = pya.Trans.from_s('r0 %s,%s' % (x,y))
    instGC1 = cell.insert(pya.CellInstArray(cell_ebeam_gc.cell_index(), t))
    t = pya.Trans.from_s('r0 %s,%s' % (x,y+127000))
    instGC2 = cell.insert(pya.CellInstArray(cell_ebeam_gc.cell_index(), t))

    # automated test label
    text = pya.Text ("opt_in_TE_1550_device_%s_MZI2" % designer_name, t)
    cell.shapes(ly.layer(ly.TECHNOLOGY['Text'])).insert(text).text_size = 5/dbu

    # Y branches:
    instY1 = connect_cell(instGC1, 'opt1', cell_ebeam_y_dream, 'opt1')
    instY1.transform(Trans(20000,0))
    instY2 = connect_cell(instGC2, 'opt1', cell_ebeam_y_dream, 'opt1')
    instY2.transform(Trans(20000,0))

    # Waveguides:

    connect_pins_with_waveguide(instGC1, 'opt1', instY1, 'opt1', waveguide_type=waveguide_type)
    connect_pins_with_waveguide(instGC2, 'opt1', instY2, 'opt1', waveguide_type=waveguide_type)
    connect_pins_with_waveguide(instY1, 'opt2', instY2, 'opt3', waveguide_type=waveguide_type)
    connect_pins_with_waveguide(instY1, 'opt3', instY2, 'opt2', waveguide_type=waveguide_type,turtle_B=[25,-90])



    # 3rd MZI, with a very long delay line
    cell_ebeam_delay = ly.create_cell('spiral_paperclip', 'EBeam_Beta',{
                            'waveguide_type':waveguide_type_delay,
                            'length':200,
                            'loops':4,
                            'flatten':True})
    x,y = 60000, 175000
    t = Trans(Trans.R0,x,y)
    instGC1 = cell.insert(CellInstArray(cell_ebeam_gc.cell_index(), t))
    t = Trans(Trans.R0,x,y+127000)
    instGC2 = cell.insert(CellInstArray(cell_ebeam_gc.cell_index(), t))

    # automated test label
    text = Text ("opt_in_TE_1550_device_%s_MZI3" % designer_name, t)
    cell.shapes(ly.layer(ly.TECHNOLOGY['Text'])).insert(text).text_size = 5/dbu

    # Y branches:
    instY1 = connect_cell(instGC1, 'opt1', cell_ebeam_y_dream, 'opt1')
    instY1.transform(Trans(20000,0))
    instY2 = connect_cell(instGC2, 'opt1', cell_ebeam_y_dream, 'opt1')
    instY2.transform(Trans(20000,0))

    # Spiral:
    instSpiral = connect_cell(instY2, 'opt2', cell_ebeam_delay, 'optA')
    instSpiral.transform(Trans(20000,0))

    # Waveguides:
    connect_pins_with_waveguide(instGC1, 'opt1', instY1, 'opt1', waveguide_type=waveguide_type)
    connect_pins_with_waveguide(instGC2, 'opt1', instY2, 'opt1', waveguide_type=waveguide_type)
    connect_pins_with_waveguide(instY1, 'opt2', instY2, 'opt3', waveguide_type=waveguide_type)
    connect_pins_with_waveguide(instY2, 'opt2', instSpiral, 'optA', waveguide_type=waveguide_type)
    connect_pins_with_waveguide(instY1, 'opt3', instSpiral, 'optB', waveguide_type=waveguide_type,turtle_B=[5,-90])



    # Zoom out
    zoom_out(cell)

    # Save with PCell details
    path = os.path.dirname(os.path.realpath(__file__))
    filename = os.path.splitext(os.path.basename(__file__))[0]
    ly.write(os.path.join(path,'..',filename+'.gds'))

    # Export for fabrication:
    file_out = export_layout(cell, path, filename, relative_path = '..', format='oas', screenshot=True)

if __name__ == "__main__":
    MZI()



"""
Run the examples as tests

by Lukas Chrostowski 2024

"""

import sys, os
filepath = os.path.dirname(os.path.realpath(__file__))
newpath = os.path.join(filepath,'../Example_scripted_layouts')
sys.path.append(newpath)


def test_examples():
    
    import MZI
    MZI.MZI()

    import Ring_resonator_sweep
    Ring_resonator_sweep.Ring_resonator_sweep()
    
    import Contra_directional_coupler_design
    Contra_directional_coupler_design.Contra_directional_coupler_design()
    
if __name__ == "__main__":
    test_examples()





#  Unit test for simulation using OPICS


def test_opics():
    import SiEPIC

    # using https://github.com/siepic/opics
    # in SiEPIC-Tools
    import SiEPIC.opics as opics
    from SiEPIC.opics.network import Network
    from SiEPIC.opics.globals import C

    import siepic_ebeam_pdk

    import opics_ebeam
    ebeam_lib = opics_ebeam

    print(f"Components: {ebeam_lib.components_list}")

    # Create a Mach-Zehnder circuit simulation
    import numpy as np
    freq = np.linspace(C * 1e6 / 1.45, C * 1e6 / 1.61, 5000)
    circuit = Network(network_id="circuit_name", f=freq)

    input_gc = circuit.add_component(ebeam_lib.GC)
    y = circuit.add_component(ebeam_lib.Y)
    wg2 = circuit.add_component(ebeam_lib.ebeam_wg_integral_1550, params=dict(wg_length=40e-6))
    wg1 = circuit.add_component(ebeam_lib.Waveguide, params={"length":15e-6})
    y2 = circuit.add_component(ebeam_lib.Y)
    output_gc = circuit.add_component(ebeam_lib.GC)

    input_gc.set_port_reference(0, "input_port")
    output_gc.set_port_reference(0, "output_port")

    circuit.connect(input_gc, 1, y, 0)
    circuit.connect(y, 1, wg1, 0)
    circuit.connect(y, 2, wg2, 0)
    circuit.connect(y2, 0, output_gc, 1)
    circuit.connect(wg1, 1, y2, 1)
    circuit.connect(wg2, 1, y2, 2)

    circuit.simulate_network()

    result = circuit.sim_result.get_data()
    wavelengths = C/circuit.f
    transmission = result['S_0_1']
    reflection = result['S_0_0']
    
    print(len(transmission))

    assert len(transmission)==5000

if __name__ == "__main__":
    test_opics()



import os
from pya import (
    PCellDeclarationHelper,
    Point,
    DPoint,
    Path,
    Polygon,
    Text,
    Box,
    LayerInfo,
    Trans,
    CellInstArray,
    Library,
)
from math import pi, acos, sqrt
from SiEPIC.extend import to_itype
from SiEPIC.utils import arc_wg_xy, get_technology_by_name
from SiEPIC.utils.layout import new_layout
from SiEPIC.scripts import zoom_out
from SiEPIC._globals import PIN_LENGTH as pin_length, Python_Env


class strip_to_slot(PCellDeclarationHelper):
    """
    The PCell declaration for the strip_to_slot.
    draft by Lukas Chrostowski july 24, 2017
    Modified by Alexander Tofini march 17, 2021
    based on https://www.osapublishing.org/oe/fulltext.cfm?uri=oe-21-16-19029&id=259920

    updated by Lukas Chrostowski, 2021/12
      - reduce the curved waveguide section, longer device for lower loss
      - simulated using EME to give 99% efficiency

    updated by Mustafa Hammood, 2025/02
      - refactored redundant code, added in-cell execution
      - added optional theta for slot curve (via *radius*)
    """

    def __init__(self):
        super(strip_to_slot, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")
        # declare parameters
        self.param("silayer", self.TypeLayer, "Si Layer", default=TECHNOLOGY["Si"])
        self.param("r", self.TypeDouble, "Radius", default=15)
        self.param("w", self.TypeDouble, "Waveguide Width", default=0.5)
        self.param("rails", self.TypeDouble, "Rails", default=0.25)
        self.param("slot", self.TypeDouble, "Slot", default=0.1)
        self.param("taper", self.TypeDouble, "Taper Length", default=15.0)
        self.param("wt", self.TypeDouble, "Terminated Curved Width", default=0.1)
        self.param(
            "offset", self.TypeDouble, "Ending Offset for Curved section", default=1.0
        )
        self.param(
            "pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"]
        )
        self.param(
            "devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"]
        )
        self.param("textl", self.TypeLayer, "Text Layer", default=LayerInfo(10, 0))

    def display_text_impl(self):
        return (
            "Strip_To_Slot(rails="
            + ("%.3f" % self.rails)
            + ", slot="
            + ("%g" % self.slot)
            + ")"
        )

    def can_create_from_shape_impl(self):
        return False

    def produce_impl(self):

        dbu = self.layout.dbu
        ly = self.layout
        shapes = self.cell.shapes

        LayerSiN = ly.layer(self.silayer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        # Convert parameters to internal units
        w = to_itype(self.w, dbu)
        r = to_itype(self.r, dbu)
        slot = to_itype(self.slot, dbu)
        rails = to_itype(self.rails, dbu)
        taper = to_itype(self.taper, dbu)
        wt = to_itype(self.wt, dbu)
        offset = to_itype(self.offset, dbu)

        # Draw the **curved arc** for the slot converter
        x = 0
        y = r + rails / 2 + w + slot
        theta = (acos((r - offset) / r) / (2 * pi)) * 360
        arc_width = sqrt(r**2 - (r - offset) ** 2)
        self.cell.shapes(LayerSiN).insert(
            arc_wg_xy(x, y, r + rails / 2 - wt / 2, wt, 270, 270 + theta)
        )

        # **Taper polygon**
        pts = [
            Point.from_dpoint(DPoint(0, w)),
            Point.from_dpoint(DPoint(0, 0)),
            Point.from_dpoint(DPoint(-taper, w - rails)),
            Point.from_dpoint(DPoint(-taper, w)),
        ]
        shapes(LayerSiN).insert(Polygon(pts))

        # **Top left half waveguide** (slot side)
        wg1 = Polygon(
            [
                Point.from_dpoint(DPoint(-taper, w + slot)),
                Point.from_dpoint(DPoint(-taper, w + slot + rails)),
                Point.from_dpoint(DPoint(0, w + slot + wt)),
                Point.from_dpoint(DPoint(0, w + slot)),
            ]
        )
        shapes(LayerSiN).insert(wg1)

        # **Bus waveguide**
        wg2 = Box(0, 0, arc_width + wt, w)
        shapes(LayerSiN).insert(wg2)

        # **Pins**
        pin_path1 = Path(
            [
                Point(-taper + pin_length, w + slot / 2),
                Point(-taper - pin_length, w + slot / 2),
            ],
            2 * rails + slot,
        )
        shapes(LayerPinRecN).insert(pin_path1)
        text1 = Text("opt1", Trans(Trans.R0, -taper, w))
        text1.text_size = 0.5 / dbu
        shapes(LayerPinRecN).insert(text1)

        pin_path2 = Path(
            [
                Point(arc_width + wt - pin_length, w / 2),
                Point(arc_width + wt + pin_length, w / 2),
            ],
            w,
        )
        shapes(LayerPinRecN).insert(pin_path2)
        text2 = Text("opt2", Trans(Trans.R0, arc_width + wt, w / 2))
        text2.text_size = 0.5 / dbu
        shapes(LayerPinRecN).insert(text2)

        # **Device recognition layer**
        dev = Box(-taper, -w / 2 - w, arc_width + wt, y - r + offset)
        shapes(LayerDevRecN).insert(dev)


class test_lib(Library):
    def __init__(self):
        tech = "EBeam"
        library = tech + "test_lib"
        self.technology = tech
        self.layout().register_pcell("strip_to_slot", strip_to_slot())
        self.register(library)


if __name__ == "__main__":
    print("Test layout for: Strip to Slot")

    if Python_Env == "Script":
        # For external Python mode, when installed using pip install siepic_ebeam_pdk
        import siepic_ebeam_pdk

    # load the test library, and technology
    t = test_lib()
    tech = t.technology

    # Create a new layout for the chip floor plan
    topcell, ly = new_layout(tech, "test", GUI=True, overwrite=True)

    # instantiate the cell
    library = tech + "test_lib"

    # Create a strip_to_slot PCell
    pcell = ly.create_cell("strip_to_slot", library, {})
    t = Trans(Trans.R0, 0, 0)
    topcell.insert(CellInstArray(pcell.cell_index(), t))

    # Display the layout in KLayout, using KLayout Package "klive", which needs to be installed in the KLayout Application
    if Python_Env == "Script":
        from SiEPIC.scripts import export_layout

        path = os.path.dirname(os.path.realpath(__file__))
        file_out = export_layout(
            topcell, path, filename="strip_to_slot", relative_path="", format="oas"
        )
        from SiEPIC.utils import klive

        klive.show(file_out, technology=tech)



import pya
from pya import DPolygon, DPoint, DBox, DTrans, CellInstArray, Trans, Library
import os
# from pya import *
from SiEPIC.utils import get_technology_by_name


class PWB_edge_coupler(pya.PCellDeclarationHelper):
    """
    The PCell declaration for the photonic wirebond edge taper

    based on Becky Lin's 
    Author: Becky Lin SiEPIC_EBeam_PDK ebeam_PWB_surface_taper cell

    by Lukas Chrostowski, 2025
    
    Waveguide choices are loaded from WAVEGUIDES.xml
     - determiners the layer/material
     - reads the waveguide width, for the port
     - The taper is constructed using a single layer.
     - does not work for compound waveguides.
        
    """

    def __init__(self):
        # Important: initialize the super class
        self.cellName = self.__class__.__name__
        super(eval(self.cellName), self).__init__()

        self.technology_name = "EBeam"
        from SiEPIC.utils import load_Waveguides_by_Tech
        self.TECHNOLOGY = get_technology_by_name(self.technology_name)

        # Load all strip waveguides
        self.waveguide_types = load_Waveguides_by_Tech(self.technology_name)

        # declare the parameters
        p = self.param("waveguide_type", self.TypeList, "Waveguide Type", default=self.waveguide_types[0]["name"])
        for wa in self.waveguide_types:
            if not "compound_waveguide" in wa:
                p.add_choice(wa["name"], wa["name"])

        # Taper
        self.param("taper_tip_width", self.TypeDouble, "Width of taper tip (microns)", default=0.13)
        #self.param("edge_offset", self.TypeDouble, "tip to edge gap (microns), negative extends past the etch", default=-10)
        self.param("taper_length", self.TypeDouble, "Waveguide taper length (microns)", default=100)
        self.param("taper_extension", self.TypeDouble, "Extension (straight) of the taper tip (microns)", default=10)

        # Alignment marker
        self.param("offset_align_tip", self.TypeDouble, "Alignment marker: offset from taper tip (microns)", default=10)
        self.param("pitch_align", self.TypeDouble, "Alignment marker: pitch of the two markers (microns)", default=50)

        # Trench
        self.param("trench_width", self.TypeDouble, "Trench width (microns)", default=100)
        self.param("trench_height", self.TypeDouble, "Trench height (microns)", default=50)
        self.param("layer_deep_etch", self.TypeLayer, "Trench layer", default=self.TECHNOLOGY["Deep Trench"])

        self.param("include_DevRec", self.TypeBoolean, "Draw the DevRec layer?", default=True)

        

    def get_waveguide_parameters(self):
        """
        Get the Waveguide Type from the PCell, and update waveguide parameters
        """

        from SiEPIC.utils import load_Waveguides_by_Tech
        # Load all waveguides
        self.TECHNOLOGY = get_technology_by_name(self.technology_name)
        self.waveguide_types = load_Waveguides_by_Tech(self.technology_name)
        # print(self.waveguide_types)

        self.waveguide_params = [
            t for t in self.waveguide_types if t["name"] == self.waveguide_type
        ]
        if not self.waveguide_params:
            raise Exception("Waveguides.XML not correctly configured")
        self.waveguide_params = self.waveguide_params[0]
        wg_params = self.waveguide_params
        wg_params2 = wg_params

        # check for compound type
        if "compound_waveguide" in wg_params:
            sm_wg = self.waveguide_params["compound_waveguide"]["singlemode"]
            mm_wg = self.waveguide_params["compound_waveguide"]["multimode"]
            wg_params = [t for t in self.waveguide_types if t["name"] == sm_wg][0]
            wg_params2 = [t for t in self.waveguide_types if t["name"] == mm_wg][0]

        # DevRec width
        if "DevRec" not in [wg["layer"] for wg in wg_params2["component"]]:
            from SiEPIC import _globals

            self.devrec = (
                max([float(wg["width"]) for wg in wg_params2["component"]])
                + _globals.WG_DEVREC_SPACE * 2
            )
        else:
            self.devrec = float(
                [f for f in wg_params2["component"] if f["layer"] == "DevRec"][0][
                    "width"
                ]
            )

        # Radius
        self.radius = float(wg_params["radius"])

        # Waveguide pin width
        self.wg_width = float(wg_params["width"])
        
        # Waveguide layer / material
        wg_layers = ['Si', 'SiN']
        for w in wg_layers:
            if any([f for f in wg_params2["component"] if f["layer"] == w]):
                self.waveguide_layer = w


    def produce_impl(self):

        # Cell origin is the waveguide port
        
        self.get_waveguide_parameters()
        wg_width = self.wg_width
        self.waveguide_layer

        # define the taper
        pts = [
            DPoint(-self.taper_length, -self.taper_tip_width/2),
            DPoint(-self.taper_length-self.taper_extension, -self.taper_tip_width/2),
            DPoint(-self.taper_length-self.taper_extension, self.taper_tip_width/2),
            DPoint(-self.taper_length, self.taper_tip_width/2),
            DPoint(0, self.wg_width/2),
            DPoint(0, -self.wg_width/2),
        ]
        # place the taper
        layer = self.layout.layer(self.TECHNOLOGY[self.waveguide_layer])
        self.cell.shapes(layer).insert(DPolygon(pts))

        # Create the pins on the input waveguide
        from SiEPIC.utils.layout import make_pin
        make_pin(
            self.cell,
            "optA",
            [0, 0],
            self.wg_width,
            self.layout.layer(self.TECHNOLOGY["PinRec"]),
            0,
        )
        
        # draw alignment marker
        pts = [
            DPoint(0, 12.4 ),
            DPoint(0, 9 ),
            DPoint(3.5, 9 ),
            DPoint(3.5, 12.5 ),
            DPoint(0.1, 12.5 ),
            DPoint(0, 12.6 ),
            DPoint(0, 16 ),
            DPoint(-3.5, 16 ),
            DPoint(-3.5, 12.5 ),
            DPoint(-0.1, 12.5 ),
        ]

        # place alignment markers
        self.cell.shapes(layer).insert(DPolygon(pts).transformed(DTrans(DTrans.R0, -self.taper_length+self.offset_align_tip,0)))
        self.cell.shapes(layer).insert(DPolygon(pts).transformed(DTrans(DTrans.R0, -self.taper_length+self.pitch_align,0)))
 
        # Trench layer
        if self.layer_deep_etch:
            box_trench = DBox (-self.taper_length, 
                               -self.trench_height/2, 
                               -self.taper_length-self.trench_width, 
                               self.trench_height/2)
            layer = self.layout.layer(self.layer_deep_etch)
            self.cell.shapes(layer).insert(box_trench)
                    
        # DevRec
        if self.include_DevRec:
            box = DBox (-self.taper_length-self.taper_extension, -17, 0, 17)
            layer = self.layout.layer(self.TECHNOLOGY['DevRec'])
            if self.layer_deep_etch:
                dbu = self.layout.dbu
                box = pya.Region(box.to_itype(dbu)) + pya.Region(box_trench.to_itype(dbu))
                box.merge()
            self.cell.shapes(layer).insert(box)

class test_lib(Library):
    def __init__(self):
        tech = "EBeam"
        library = "test_lib"
        self.technology = tech
        self.layout().register_pcell("PWB_edge_coupler", PWB_edge_coupler())
        self.register(library)


if __name__ == "__main__":
    print("Test layout for: PWB_edge_coupler")

    from SiEPIC.utils.layout import new_layout
    from SiEPIC.scripts import zoom_out

    from SiEPIC._globals import Python_Env
    if Python_Env == 'Script':
        # For external Python mode, when installed using pip install siepic_ebeam_pdk
        import siepic_ebeam_pdk

    # load the test library, and technology
    t = test_lib()
    tech = t.technology

    # Create a new layout for the chip floor plan
    topcell, ly = new_layout(tech, "test", GUI=True, overwrite=True)
    # floorplan(topcell, 100e3, 100e3)

    # instantiate the cell
    library = "test_lib"

    # Create spirals for all the types of waveguides
    from SiEPIC.utils import load_Waveguides_by_Tech

    waveguide_types = load_Waveguides_by_Tech(tech)

    y = 0
    pcell = ly.create_cell(
        "PWB_edge_coupler",
        library,
        {
            "waveguide_type": waveguide_types[-1]["name"],
            "include_DevRec":True
        },
    )
    t = Trans(Trans.R0, -pcell.bbox().width(), y)
    inst = topcell.insert(CellInstArray(pcell.cell_index(), t))
    y += pcell.bbox().height() + 10e3
    
    pcell = ly.create_cell(
        "PWB_edge_coupler",
        library,
        {
            "waveguide_type": waveguide_types[-1]["name"],
            "include_DevRec":False
        },
    )
    t = Trans(Trans.R0, -pcell.bbox().width(), y)
    inst = topcell.insert(CellInstArray(pcell.cell_index(), t))
    y += pcell.bbox().height() + 10e3

    pcell = ly.create_cell(
        "PWB_edge_coupler",
        library,
        {
            "waveguide_type": waveguide_types[-1]["name"],
            "layer_deep_etch":None,
            "include_DevRec":True,
        },
    )
    t = Trans(Trans.R0, -pcell.bbox().width(), y)
    inst = topcell.insert(CellInstArray(pcell.cell_index(), t))

    topcell.show()

    # # Display the layout in KLayout, using KLayout Package "klive", which needs to be installed in the KLayout Application
    # if Python_Env == 'Script':
    #     from SiEPIC.scripts import export_layout
    #     path = os.path.dirname(os.path.realpath(__file__))
    #     file_out = export_layout (topcell, path, filename='PWB_edge_coupler', relative_path='', format='oas')
    #     from SiEPIC.utils import klive
    #     klive.show(file_out, technology=tech)



# $description: Double-bus ring resonator sweep (EBeam)
# $show-in-menu
# $group-name: Examples_EBeam
# $menu-path: siepic_menu.exlayout.begin

"""
Scripted layout for ring resonators using SiEPIC-Tools
in the SiEPIC-EBeam-PDK "EBeam" technology

by Lukas Chrostowski, 2020-2023
"""

def Ring_resonator_sweep():
    print("SiEPIC_EBeam_PDK: Example - Ring_resonator_sweep.py")
    import pya
    from pya import Trans, CellInstArray, Text

    import SiEPIC
    from SiEPIC._globals import Python_Env
    from SiEPIC.scripts import zoom_out, export_layout
    import os

    if Python_Env == "Script":
        try:
            # For external Python mode, when installed using pip install siepic_ebeam_pdk
            import siepic_ebeam_pdk
        except:
            # Load the PDK from a folder, e.g, GitHub, when running externally from the KLayout Application
            import os, sys

            path_GitHub = os.path.expanduser("~/Documents/GitHub/")
            sys.path.append(os.path.join(path_GitHub, "SiEPIC_EBeam_PDK/klayout"))
            import siepic_ebeam_pdk

    tech_name = "EBeam"


    # Example layout function
    def dbl_bus_ring_res():
        # Import functions from SiEPIC-Tools
        from SiEPIC.extend import to_itype
        from SiEPIC.scripts import connect_cell, connect_pins_with_waveguide
        from SiEPIC.utils.layout import new_layout, floorplan

        # Create a layout for testing a double-bus ring resonator.
        # uses:
        #  - the SiEPIC EBeam Library
        # creates the layout in the presently selected cell
        # deletes everything first

        # Configure parameter sweep
        pol = "TE"
        if pol == "TE":
            sweep_radius = [3, 5, 5, 5, 10, 10, 10, 10, 10]
            sweep_gap = [0.07, 0.07, 0.08, 0.09, 0.07, 0.08, 0.09, 0.10, 0.11]
            x_offset = 67
            wg_bend_radius = 5
        else:
            sweep_radius = [30, 30, 30, 30, 30, 30, 30, 30]
            sweep_gap = [0.25, 0.3, 0.35, 0.4, 0.45, 0.5, 0.55, 0.60]
            x_offset = 67
            wg_bend_radius = 15

        wg_width = 0.5

        """
        Create a new layout using the EBeam technology,
        with a top cell
        and Draw the floor plan
        """
        cell, ly = new_layout(tech_name, "top", GUI=True, overwrite=True)
        floorplan(cell, 605e3, 410e3)

        from packaging import version

        if version.parse(SiEPIC.__version__) < version.parse("0.5.1"):
            pya.MessageBox.warning(
                "Errors",
                "This example requires SiEPIC-Tools version 0.5.1 or greater.",
                pya.MessageBox.Ok,
            )

        # Layer mapping:
        LayerSiN = ly.layer(ly.TECHNOLOGY["Si"])
        fpLayerN = cell.layout().layer(ly.TECHNOLOGY["FloorPlan"])
        TextLayerN = cell.layout().layer(ly.TECHNOLOGY["Text"])

        # Create a sub-cell for our Ring resonator layout
        top_cell = cell
        dbu = ly.dbu
        cell = cell.layout().create_cell("RingResonator")
        if pol == "TE":
            t = Trans(Trans.R0, 40 / dbu, 12 / dbu)
        else:
            # rotate the layout since EBeam TM grating couplers have an angle that is negative
            t = Trans(Trans.R180, 560 / dbu, 393 / dbu)

        # place the cell in the top cell
        top_cell.insert(CellInstArray(cell.cell_index(), t))

        # Import cell from the SiEPIC EBeam Library
        cell_ebeam_gc = ly.create_cell("GC_%s_1550_8degOxide_BB" % pol, "EBeam")
        # get the length of the grating coupler from the cell
        gc_length = cell_ebeam_gc.bbox().width() * dbu
        # spacing of the fibre array to be used for testing
        GC_pitch = 127

        # Loop through the parameter sweep
        for i in range(len(sweep_gap)):
            # place layout at location:
            if i == 0:
                x = 0
            else:
                # next device is placed at the right-most element + length of the grating coupler
                x = inst_dc2.bbox().right * dbu + gc_length + 1

            # get the parameters
            r = sweep_radius[i]
            g = sweep_gap[i]

            # Grating couplers, Ports 0, 1, 2, 3 (from the bottom up)
            instGCs = []
            for i in range(0, 4):
                t = Trans(Trans.R0, to_itype(x, dbu), i * 127 / dbu)
                instGCs.append(cell.insert(CellInstArray(cell_ebeam_gc.cell_index(), t)))

            # Label for automated measurements, laser on Port 2, detectors on Ports 1, 3, 4
            t = Trans(Trans.R90, to_itype(x, dbu), to_itype(GC_pitch * 2, dbu))
            text = Text(
                "opt_in_%s_1550_device_RingDouble%sr%sg%s"
                % (pol.upper(), pol.upper(), r, int(round(g * 1000))),
                t,
            )
            text.halign = 1
            cell.shapes(TextLayerN).insert(text).text_size = 5 / dbu

            """
            # Label for automated measurements, laser on Port 1, detectors on Ports 2, 3
            # this will cause an error in the verification since the 4th detector is not connected
            t = Trans(Trans.R0, to_itype(x,dbu), to_itype(GC_pitch*3,dbu))
            text = Text ("opt_in_%s_1550_device_RingDouble%sr%sg%sB" % (pol.upper(), pol.upper(),r,int(round(g*1000))), t)
            text.halign = 1
            cell.shapes(TextLayerN).insert(text).text_size = 5/dbu
            """

            # Ring resonator from directional coupler PCells
            cell_dc = ly.create_cell(
                "ebeam_dc_halfring_straight",
                "EBeam",
                {"r": r, "w": wg_width, "g": g, "bustype": 0},
            )
            y_ring = GC_pitch * 3 / 2
            # first directional coupler
            t1 = Trans(Trans.R270, to_itype(x + wg_bend_radius, dbu), to_itype(y_ring, dbu))
            inst_dc1 = cell.insert(CellInstArray(cell_dc.cell_index(), t1))
            # add 2nd directional coupler, snapped to the first one
            inst_dc2 = connect_cell(inst_dc1, "pin2", cell_dc, "pin4")

            # Create paths for waveguides, with the type defined in WAVEGUIDES.xml in the PDK
            waveguide_type = "Strip TE 1550 nm, w=500 nm"

            # GC1 to bottom-left of ring pin3
            connect_pins_with_waveguide(
                instGCs[1], "opt1", inst_dc1, "pin3", waveguide_type=waveguide_type
            )

            # GC2 to top-left of ring pin1
            connect_pins_with_waveguide(
                instGCs[2], "opt1", inst_dc1, "pin1", waveguide_type=waveguide_type
            )

            # GC0 to top-right of ring
            connect_pins_with_waveguide(
                instGCs[0], "opt1", inst_dc2, "pin1", waveguide_type=waveguide_type
            )

            # GC3 to bottom-right of ring
            connect_pins_with_waveguide(
                instGCs[3], "opt1", inst_dc2, "pin3", waveguide_type=waveguide_type
            )

        # Introduce an error, to demonstrate the Functional Verification
        inst_dc2.transform(Trans(1000, -1000))

        return ly, cell


    ly, cell = dbl_bus_ring_res()

    # Zoom out
    zoom_out(cell)

    # Save
    path = os.path.dirname(os.path.realpath(__file__))
    filename = "Test_structures_ring_resonators"
    file_out = export_layout(
        cell, path, filename, relative_path="", format="oas", screenshot=True
    )

    from SiEPIC.verification import layout_check

    print("SiEPIC_EBeam_PDK: example_Ring_resonator_sweep.py - verification")

    file_lyrdb = os.path.join(path, filename + ".lyrdb")
    num_errors = layout_check(cell=cell, verbose=False, GUI=True, file_rdb=file_lyrdb)

    if Python_Env == "Script":
        from SiEPIC.utils import klive

        klive.show(file_out, lyrdb_filename=file_lyrdb, technology=tech_name)

    print("SiEPIC_EBeam_PDK: example_Ring_resonator_sweep.py - done")

if __name__ == "__main__":
    Ring_resonator_sweep()
    


import pya
from SiEPIC.utils import get_technology_by_name
from pya import *


class strip_to_slot_SiN(pya.PCellDeclarationHelper):
    """
    Simple strip to sloted taper for the SiN plaform
    draft by Bobby Zou Oct 24, 2024
    """

    def __init__(self):
        super(strip_to_slot_SiN, self).__init__()
        TECHNOLOGY = get_technology_by_name("EBeam")
        
        # declare the parameters
        self.param("layer", self.TypeLayer, "Layer", default = TECHNOLOGY['SiN'])
        self.param("w", self.TypeDouble, "Waveguide Width", default=1)
        self.param("rails", self.TypeDouble, "Rails", default=0.5)
        self.param("slot", self.TypeDouble, "Slot", default=0.12)
        self.param("length", self.TypeDouble, "Length", default=100)
        self.param("interfaceGap", self.TypeDouble, "Interface Slot Gap", default = 0.3)

        self.param("pinrec", self.TypeLayer, "PinRec Layer", default=TECHNOLOGY["PinRec"])
        self.param("devrec", self.TypeLayer, "DevRec Layer", default=TECHNOLOGY["DevRec"])

    def display_text_impl(self):
        # Provide a descriptive text for the cell
        return (
            "Strip_to_Slot_SiN(rails="
            + ("%.3f" % self.rails)
            + ",slot="
            + ("%g" % (self.slot))
            + ")"
            + ",width="
            + ("%g" % self.w)
        )
    
    def can_create_from_shape_impl(self):   
        return False
    

    def produce_impl(self):
        # This is the main part of the implementation: create the layout

        from SiEPIC.extend import to_itype
        from SiEPIC.utils import arc_wg_xy

        # fetch the parameters
        dbu = self.layout.dbu
        ly = self.layout
        cell = self.cell
        shapes = self.cell.shapes

        LayerSiN = ly.layer(self.layer)
        LayerPinRecN = ly.layer(self.pinrec)
        LayerDevRecN = ly.layer(self.devrec)

        # Use to_itype() to prevent rounding errors
        w = to_itype(self.w, dbu)
        rails = to_itype(self.rails, dbu)
        slot = to_itype(self.slot, dbu)
        length = to_itype(self.length, dbu)
        interfaceGap = to_itype(self.interfaceGap, dbu)

        # Create the strip to top 1/2 waveguide
        pts = []
        pts.append(Point.from_dpoint(DPoint(0, -w/2)))
        pts.append(Point.from_dpoint(DPoint(0, w/2)))
        pts.append(Point.from_dpoint(DPoint(length, slot/2 + rails)))
        pts.append(Point.from_dpoint(DPoint(length, slot/2)))
        
        polygon = Polygon(pts)
        shapes(LayerSiN).insert(polygon)

        # Create the bottom 1/2 waveguide
        pts = []
        pts.append(Point.from_dpoint(DPoint(0, -w/2 - interfaceGap - 0.12 /dbu)))
        pts.append(Point.from_dpoint(DPoint(0, -w/2 - interfaceGap)))
        pts.append(Point.from_dpoint(DPoint(length, -slot/2)))
        pts.append(Point.from_dpoint(DPoint(length, -slot/2 - rails)))
        
        polygon = Polygon(pts)
        shapes(LayerSiN).insert(polygon)

        # PinRec Left
        shapes(LayerPinRecN).insert(
            pya.Path(
                [
                    pya.Point(0.05 / dbu, 0),
                    pya.Point(-0.05 / dbu, 0),
                ],
                w,
            )
        )
        shapes(LayerPinRecN).insert(
            pya.Text("opt1", pya.Trans(pya.Trans.R0, 0, 0))
        ).text_size = 0.5 / dbu

        # PinRec Right
        shapes(LayerPinRecN).insert(
            pya.Path(
                [
                    pya.Point(length-0.05 / dbu, 0),
                    pya.Point(length+0.05 / dbu, 0),
                ],
                2 * rails + slot,
            )
        )
        shapes(LayerPinRecN).insert(
            pya.Text("opt2", pya.Trans(pya.Trans.R0, length, 0))
        ).text_size = 0.5 / dbu

        # Create the device recognition layer -- make it 1 * wg_width away from the waveguides.
        dev = pts = []
        pts.append(Point.from_dpoint(DPoint(0, -w/2 - interfaceGap - 0.12 /dbu - w)))
        pts.append(Point.from_dpoint(DPoint(0, w/2 + w)))
        pts.append(Point.from_dpoint(DPoint(length, slot/2 + rails + w)))
        pts.append(Point.from_dpoint(DPoint(length, -slot/2 - rails - w)))
        polygon = Polygon(pts)
        shapes(LayerDevRecN).insert(polygon)


# $description: Contra-directional coupler design (EBeam)
# $show-in-menu
# $group-name: Examples_EBeam
# $menu-path: siepic_menu.exlayout.begin

def Contra_directional_coupler_design():
    """

    Create a complete layout for a contra-directional coupler circuit

    using KLayout & SiEPIC-Tools scripted layout

    usage:
    - the SiEPIC EBeam Library
    - Uncomment path_to_waveguide to convert everything to waveguides, or just do it on your own from SiEPIC Tools menu.
    - Run circuit simulations either using OPICS or INTERCONNECT

    Script will generate an series of devices with the selected parameters, and generate GCs for testing with automated measurement labels

    Author: Lukas Chrostowski, Mustafa Hammood

    March 2024
    """

    print("SiEPIC_EBeam_PDK: Example - Example - Contra-directional coupler designs.py")

    # Import KLayout-Python API
    import pya
    from pya import Trans, CellInstArray, Text

    import SiEPIC
    from SiEPIC._globals import Python_Env
    from SiEPIC.scripts import zoom_out, export_layout
    import os

    if Python_Env == "Script":
        try:
            # For external Python mode, when installed using pip install siepic_ebeam_pdk
            import siepic_ebeam_pdk
        except:
            # Load the PDK from a folder, e.g, GitHub, when running externally from the KLayout Application
            import os, sys

            path_GitHub = os.path.expanduser("~/Documents/GitHub/")
            sys.path.append(os.path.join(path_GitHub, "SiEPIC_EBeam_PDK/klayout"))
            import siepic_ebeam_pdk

    tech_name = "EBeam"

    # Import from SiEPIC-Tools
    from SiEPIC.extend import to_itype
    from SiEPIC.scripts import connect_pins_with_waveguide, connect_cell

    from packaging import version

    if version.parse(SiEPIC.__version__) < version.parse("0.5.1"):
        pya.MessageBox.warning(
            "Errors",
            "This example requires SiEPIC-Tools version 0.5.1 or greater.",
            pya.MessageBox.Ok,
        )


    # define layout parameters in the class below
    # ALL distance parameters are in microns unless specified otherwise
    class parameters:
        """Define the circuit layout parameters"""

        # Default contraDC PCell parameters
        component_contraDC = "contra_directional_coupler"
        libname = "EBeam"
        N = 1000
        period = 0.316
        g = 100
        w1 = 560
        w2 = 440
        dW1 = 0.05
        dW2 = 0.025
        sine = 0
        a = 2.7

        # routing and device placement parameters
        x_offset = 170  # spacing between grating couplers columns
        wg_bend_radius = 35  # waveguide routes bend radius
        device_spacing = 7.3  # spacing between devices
        wg_pitch = (
            5  # spacing between waveguides ( keep > 2 microns to minimize cross coupling)
        )
        pol = "TE"
        waveguide_type = "Strip TE 1550 nm, w=500 nm"
        gc_pitch = 127  # spacing between grating couplers array (for a single device)
        cdc_offset = 20
        name = "contraDC"


    params = parameters()


    def import_fixed_cells(ly):
        """Import the fixed cells from the library, and add them to the layout"""
        params.cell_gc = ly.create_cell(
            "ebeam_gc_%s1550" % params.pol.lower(), "EBeam"
        )  # .cell_index()
        # params.cell_y = ly.create_cell("ebeam_y_1550", "EBeam") # .cell_index()

        params.gc_length = 41  # Length of a grating coupler cell
        params.gc_height = 30  # Height of a grating coupler cell


    def ebeam_c_te_mux_1ch_standard_1543nm(
        topcell,
        x_pos=0,
        y_pos=0,
        N=params.N,
        period=params.period,
        g=0.1,
        w1=0.56,
        w2=0.44,
        dW1=0.048,
        dW2=0.024,
        sine=0,
        a=2.7,
        wg_width=0.5,
        pol=params.pol,
    ):
        """Create a layout of a contradirectional coupler, with grating couplers
        This is a known-good design by Mustafa Hammood,
        having been fabricated by Applied Nanotools, and tested at UBC
        It is in the C-band and is used for the Process Control Monitor (PCM) structure (C-PCM).
        Works well when cascaded.
        Strip waveguides
        Bandwidth ~6 nm;	Wavelength 1543 nm
        Data: single stage, cascaded, exists on every PCM
        function by Lukas Chrostowski
        """

        # Create a sub-cell for our contraDC layout
        cell = topcell.layout().create_cell("contraDC_GCarray")

        # place the cell in the top cell
        t = Trans(Trans.R0, x_pos, y_pos)
        topcell.insert(CellInstArray(cell.cell_index(), t))
        ly = topcell.layout()

        # Grating couplers, Ports 1, 2, 3, 4 (top-down)
        instGCs = []
        for i in range(4):
            t = Trans(
                Trans.R0,
                to_itype(params.gc_length, ly.dbu),
                to_itype(params.gc_height / 2 + i * params.gc_pitch, ly.dbu),
            )
            instGCs.append(cell.insert(CellInstArray(params.cell_gc.cell_index(), t)))
        t = Trans(
            Trans.R0,
            to_itype(params.gc_length, ly.dbu),
            to_itype(params.gc_height / 2 + params.gc_pitch * 2, ly.dbu),
        )
        text = Text(
            "opt_in_%s_1550_device_%s%dN%dperiod%dg%dwa%dwb%ddwa%ddwb%dsine%sa"
            % (
                pol,
                params.name,
                N,
                1000 * period,
                1000 * g,
                1000 * w1,
                1000 * w2,
                1000 * dW1,
                1000 * dW2,
                sine,
                a,
            ),
            t,
        )
        # text = Text ("opt_in_TE_1550_device_contraDC1"
        shape = cell.shapes(ly.layer(ly.TECHNOLOGY["Text"])).insert(text)
        shape.text_size = 1.5 / ly.dbu

        # contraDC PCell
        pcell = ly.create_cell(
            params.component_contraDC,
            params.libname,
            {
                "sbend": 1,
                "number_of_periods": N,
                "grating_period": period,
                "gap": g,
                "wg1_width": w1,
                "wg2_width": w2,
                "corrugation_width1": dW1,
                "corrugation_width2": dW2,
                "sinusoidal": sine,
                "index": a,
            },
        )
        if not pcell:
            raise Exception(
                "Cannot find cell %s in library %s."
                % (params.component_contraDC, params.libname)
            )
        t = Trans(
            Trans.R90,
            to_itype(params.gc_length + params.cdc_offset, ly.dbu),
            to_itype(params.gc_height / 2 + params.gc_pitch * 0.2, ly.dbu),
        )
        instCDC = cell.insert(CellInstArray(pcell.cell_index(), t))

        # Waveguides:
        connect_pins_with_waveguide(
            instGCs[3], "opt1", instCDC, "opt3", waveguide_type=params.waveguide_type
        )
        connect_pins_with_waveguide(
            instGCs[2],
            "opt1",
            instCDC,
            "opt4",
            waveguide_type=params.waveguide_type,
            turtle_A=[5, 90, 30, -90],
            turtle_B=[5, 90],
        )
        connect_pins_with_waveguide(
            instGCs[1],
            "opt1",
            instCDC,
            "opt2",
            waveguide_type=params.waveguide_type,
            turtle_A=[5, -90, 30, 90],
            turtle_B=[5, -90],
        )
        connect_pins_with_waveguide(
            instGCs[0], "opt1", instCDC, "opt1", waveguide_type=params.waveguide_type
        )

        return cell


    def layout_contraDC_circuits(newlayout=True):
        """
        Generates contraDC circuits.
        Either create a new layout using the EBeam technology,
            newlayout = True
        or delete everything in the present layout
            newlayout = False
        """

        from SiEPIC.utils.layout import new_layout, floorplan
        from SiEPIC.utils import select_paths, get_layout_variables

        """
        Create a new layout using the EBeam technology, with a top cell
        """
        topcell, ly = new_layout(
            tech_name, "SiEPIC_EBeam_contraDC_circuits", GUI=True, overwrite=not newlayout
        )

        # Import the grating couplers, and other fixed cells
        import_fixed_cells(ly)

        # create a floor plan
        # 605 x 410 microns is the space allocation for the edX course and openEBL
        # https://siepic.ca/openebl/
        floorplan(topcell, 605e3, 410e3)

        # Create the contraDC circuits
        ebeam_c_te_mux_1ch_standard_1543nm(topcell)

        # Zoom out
        zoom_out(topcell)

        # Save
        path = os.path.dirname(os.path.realpath(__file__))
        filename = "openEBL_Contra-directional-coupler-design"
        file_out = export_layout(
            topcell, path, filename, relative_path="..", format="oas", screenshot=False
        )

        from SiEPIC.verification import layout_check

        print("SiEPIC_EBeam_PDK: - verification")

        file_lyrdb = os.path.join(path, filename + ".lyrdb")
        num_errors = layout_check(
            cell=topcell, verbose=False, GUI=True, file_rdb=file_lyrdb
        )

        if Python_Env == "Script":
            from SiEPIC.utils import klive

            klive.show(file_out, lyrdb_filename=file_lyrdb, technology=tech_name)

        if num_errors == 0:
            print("Functional verification: Passed with 0 errors")
        else:
            print("Functional verification: Failed with %s errors" % num_errors)


    layout_contraDC_circuits(newlayout=False)


    # All done!

if __name__ == "__main__":
    Contra_directional_coupler_design()


'''
--- Benchmarking: MZI ---
  
by Lukas Chrostowski, 2024
 

Example simple script to
 - create a new layout with a top cell
 - create many MZI variants
 - export to OASIS for submission to fabrication

using SiEPIC-Tools function including connect_pins_with_waveguide and connect_cell

Use instructions:

Run in Python, e.g., VSCode

pip install required packages:
 - klayout, SiEPIC, siepic_ebeam_pdk, numpy

'''

import time
# Start timer
start_time = time.time()

top_cell_name = 'benchmark_MZI'
export_type = 'static'  # static: for fabrication, PCell: include PCells in file

import pya
from pya import *

import SiEPIC
from SiEPIC._globals import Python_Env
from SiEPIC.scripts import connect_cell, connect_pins_with_waveguide, zoom_out, export_layout
from SiEPIC.utils.layout import new_layout, floorplan
from SiEPIC.extend import to_itype
from SiEPIC.verification import layout_check

import os

if Python_Env == 'Script':
    # For external Python mode, when installed using pip install siepic_ebeam_pdk
    import siepic_ebeam_pdk

# Calculate and print execution time
execution_time = time.time() - start_time
print(f"Execution time (load PDK and Py modules): {execution_time} seconds")


tech_name = 'EBeam'

'''
Create a new layout using the EBeam technology,
with a top cell
and Draw the floor plan
'''    
cell, ly = new_layout(tech_name, top_cell_name, GUI=True, overwrite = True)
# floorplan(cell, 605e3, 410e3)

dbu = ly.dbu

from SiEPIC.scripts import connect_pins_with_waveguide, connect_cell
waveguide_type='Strip TE 1550 nm, w=500 nm'

# Load cells from library
cell_ebeam_y = ly.create_cell('ebeam_y_1550', tech_name)

# Array of MZI, and spacing:
n_x, n_y = 10, 10
dx, dy = 100e3, 100e3
# Each MZI 
dx_mzi = 50e3
dl_ij = 0.5
for i in range(n_x):
    for j in range(n_y):
        
        # Y branches:
        # Version 1: place it at an absolute position:
        t = Trans(dx*i,dy*j)
        instY1 = cell.insert(CellInstArray(cell_ebeam_y.cell_index(), t))
        t = Trans(Trans.R180,dx*i+dx_mzi,dy*j)
        instY2 = cell.insert(CellInstArray(cell_ebeam_y.cell_index(), t))

        # Waveguides:
        connect_pins_with_waveguide(instY1, 'opt2', instY2, 'opt3', waveguide_type=waveguide_type,turtle_A=[5,90, 0, 90], turtle_B=[5,-90, 10+dl_ij*(i*n_y+j),90])
        connect_pins_with_waveguide(instY1, 'opt3', instY2, 'opt2', waveguide_type=waveguide_type,turtle_A=[5,-90, 0, -90], turtle_B=[5,90, 10,-90])

# Zoom out
zoom_out(cell)

# Export for fabrication, removing PCells
path = os.path.dirname(os.path.realpath(__file__))
filename, extension = os.path.splitext(os.path.basename(__file__))
if export_type == 'static':
    file_out = export_layout(cell, path, filename, relative_path = '..', format='oas', screenshot=True)
else:
    file_out = os.path.join(path,'..',filename+'.oas')
    ly.write(file_out)


# Calculate and print execution time
execution_time = time.time() - start_time
print(f"Execution time (generate and save GDS): {execution_time} seconds")

# Verify
Verify = False
if Verify:
    file_lyrdb = os.path.join(path,filename+'.lyrdb')
    num_errors = layout_check(cell = cell, verbose=False, GUI=True, file_rdb=file_lyrdb)
    print('Number of errors: %s' % num_errors)
    # Calculate and print execution time
    execution_time = time.time() - start_time
    print(f"Execution time (generate, verify): {execution_time} seconds")
else:
    file_lyrdb = None
    
# Display the layout in KLayout, using KLayout Package "klive", which needs to be installed in the KLayout Application
if Python_Env == 'Script':
    from SiEPIC.utils import klive
    klive.show(file_out, lyrdb_filename=file_lyrdb, technology=tech_name)

    # Calculate and print execution time
    execution_time = time.time() - start_time
    print(f"Execution time (total, display with KLive): {execution_time} seconds")
